/*
 * g3spot <infile> [<outfile>]
 *
 * Set the PPC CPU sub-type to 0 in the given FAT binary
 * Runs on PPC or x86.
 *
 * (c) Simon Kilvington, 2009
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <mach-o/fat.h>
#include <mach-o/loader.h>

void fatal(char *, ...);

int
main(int argc, char *argv[])
{
	char *in_name;
	char *out_name;
	FILE *in_file;
	FILE *out_file;
	long size;
	unsigned char *buf;
	struct fat_header *fat_hdr;
	struct fat_arch *arch;
	struct mach_header *mach_hdr;
	unsigned int i;

	if(argc != 2 && argc != 3)
		fatal("Syntax: %s <in_file> [<out_file>]", argv[0]);

	in_name = argv[1];
	out_name = (argc == 2) ? in_name : argv[2];

	if((in_file = fopen(in_name, "r")) == NULL)
		fatal("Unable to read %s: %s", in_name, strerror(errno));

	/* find out how big it is */
	fseek(in_file, 0, SEEK_END);
	size = ftell(in_file);
	rewind(in_file);

	printf("%s: %d bytes\n", in_name, size);

	/* read it all into memory */
	if((buf = malloc(size)) == NULL)
		fatal("Out of memory");
	if(fread(buf, 1, size, in_file) != size)
		fatal("Error trying to read %s: %s", in_name, strerror(errno));
	if(fclose(in_file))
		fatal("Error trying to close %s: %s", in_name, strerror(errno));

	/* read the FAT header */
	fat_hdr = (struct fat_header *) buf;
	printf("FAT Magic: 0x%08x\n", fat_hdr->magic);
	if(fat_hdr->magic != ntohl(FAT_MAGIC))
		fatal("Wrong magic number (expecting 0x%08x)", FAT_MAGIC);
	printf("N arch: %d\n", ntohl(fat_hdr->nfat_arch));

	/* find the PPC architecture */
	for(i=0; i<ntohl(fat_hdr->nfat_arch); i++)
	{
		arch = (struct fat_arch *) (buf + sizeof(struct fat_header) + (i * sizeof(struct fat_arch)));
		printf("Archtecture %d:\n", i + 1);
		printf("CPU type: %d\n", ntohl(arch->cputype));
		printf("CPU sub-type: %d\n", ntohl(arch->cpusubtype));
		/* do we need to fix it */
		if(arch->cputype == ntohl(CPU_TYPE_POWERPC) && arch->cpusubtype != CPU_SUBTYPE_POWERPC_ALL)
		{
			printf("*** Changing to CPU sub-type 0\n");
			arch->cpusubtype = CPU_SUBTYPE_POWERPC_ALL;
			/* change the Mach-O header too */
			mach_hdr = (struct mach_header *) (buf + ntohl(arch->offset));
			printf("Mach magic: 0x%08x\n", mach_hdr->magic);
			if(mach_hdr->magic != ntohl(MH_MAGIC))
				fatal("Wrong magic number (expecting 0x%08x)", MH_MAGIC);
			printf("CPU type: %d\n", ntohl(mach_hdr->cputype));
			printf("CPU sub-type: %d\n", ntohl(mach_hdr->cpusubtype));
			printf("*** Changing to CPU sub-type 0\n");
			mach_hdr->cpusubtype = CPU_SUBTYPE_POWERPC_ALL;
		}
	}

	printf("Saving as %s\n", out_name);
	if((out_file = fopen(out_name, "w")) == NULL)
		fatal("Unable to write %s: %s", out_name, strerror(errno));
	if(fwrite(buf, 1, size, out_file) != size)
		fatal("Error trying to write %s: %s", out_name, strerror(errno));
	if(fclose(out_file))
		fatal("Error trying to close %s: %s", out_name, strerror(errno));

	return EXIT_SUCCESS;
}

void
fatal(char *msg, ...)
{
	va_list ap;

	va_start(ap, msg);
	vfprintf(stderr, msg, ap);
	fprintf(stderr, "\n");
	va_end(ap);

	exit(EXIT_FAILURE);
}

